/*
 *		b t r e e  _ c
 *
 *	This contains routines that are specific
 *	to the implementation of binary trees
 *
 *
 * (c) Copyright 1996-1999 David J. Walker
 *	   Email:  dave.walker@icl.com
 *
 *	   Permission to copy and/or distribute granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they
 *		   arise from defects in it.
 *	   3). Altered version must not be represented as being the
 *		   original software.
 *
 *
 *	AMENDMENT HISTORY
 *	~~~~~~~~~~~~~~~~~
 *	12 May 96	DJW   - First version
 *
 *	10 May 99	DJW   - Removed 'listnode' parameter from 'AddNode' routine as it
 *						was superflous and amended routine accordingly.
 *
 *	14 May 99	DJW   - Updated to handle duplicates.
 */

#include <liblist.h>
#include "liblistx.h"


/*======================================================= BTREE_ADDNODE */
static
int 	BTREE_AddNode	(list_t list, _btree_t newnode)
/*		~~~~~~~~~~~~
 *	This is a routine that tries to add a node starting at the
 *	top of the tree.	It could be written more compactly
 *	by using recursion, but this adds significant extra runtime
 *	overhead.
 *----------------------------------------------------------------------*/
{
#ifdef LIBTEST
	_btreeh_t	btreelist = (_btreeh_t)list;
#else
#define btreelist ((_btreeh_t)list)
#endif
	register  _btree_t	listnode = ((_btreeh_t)list)->nodes;
	register int reply;
	register int duplicates;
	/*
	 *	If the list is empty, then
	 *	this is a special case.
	 */

	if (listnode == NULL)
	{
		((_btreeh_t)list)->nodes = newnode;
		return LIST_ERROR_NONE;
	}


	/*
	 *	List not empty.
	 *	This is the part where recursion
	 *	could help make more compact code.
	 */
	duplicates = btreelist->listclass->duplicates;
	for ( ; ;)
	{
		/*
		 *	Handle less than case
		 */
		while ((reply = _LIST_Compare (list, (node_t)newnode, (node_t)listnode)) <= 0)
		{
			if (reply == 0)
			{
				/*
				 *	If duplicates not allowed
				 *	then return error code
				 */
				if (duplicates == LIST_DUPLICATES_NONE)
				{
					return LIST_ERROR_DUPLICATE;
				}
				/*
				 *	If duplciates are allowed, but we
				 *	are adding new ones after existing
				 *	copies then treat as greater case
				 */
				if (duplicates == LIST_DUPLICATES_AFTER)
				{
					break;
				}
			}

			if (listnode->left == NULL)
			{
				listnode->left = newnode;
				newnode->up = listnode;
				return LIST_ERROR_NONE;
			}
			listnode = listnode->left;
		}
		/*
		 *	Handle greater than case
		 */
		while (_LIST_Compare (list, (node_t)newnode, (node_t)listnode) >= 0)
		{
			if (listnode->right == NULL)
			{
				listnode->right = newnode;
				newnode->up = listnode;
				return LIST_ERROR_NONE;
			}
			listnode = listnode->right;
		}
	}
}


/*========================================================= BTREE_AFTER */
static
int 	BTREE_After 	(list_t list, node_t newnode, node_t oldnode)
/*		~~~~~~~~~~~
 *	Add a node after the given node.
 *-----------------------------------------------------------------------*/
{
#ifdef LIBTEST
	_btree_t	_newnode = (_btree_t)newnode;
#else
#define _newnode  ((_btree_t)newnode)
#endif
	(void)oldnode;		/* Stop compiler warning about unused parameter */
	LIBDEBUG ("BTREE_After", \
		   "'newnode' appears to already be linked in to a list", \
		   _newnode->up == NULL && _newnode->left == NULL && _newnode->right == NULL)

	return BTREE_AddNode (list, _newnode);
}
#undef _newnode
#undef _oldnode


/*=========================================================== BTREE_ADD */
static
int 	BTREE_Add		(list_t list, node_t node)
/*		~~~~~~~~~
 *	Add a node at the correct point.
 *----------------------------------------------------------------------*/
{
#ifdef LIBTEST
	/*
	 *	When testing library assign parameters to
	 *	local variable to help with debugging.
	 */
	_btree_t	_node = (_btree_t)node;
#else
	/*
	 *	For live code use casts (more effecient)
	 */
#define _node ((_btree_t)node)
#endif /* LIBTEST */

	LIBDEBUG ("BTREE_Add", \
		   "'node' appears to already be linked in to a list", \
		   _node->up == NULL && _node->left == NULL && _node->right == NULL)

	return BTREE_AddNode(list, _node);
}
#undef _list
#undef _node



/*========================================================================
 *
 *	This is the list of vectored functions
 *	that are used to support this list type.
 *
 *------------------------------------------------------------------------*/

static struct _listmaster master =
		{
		&_LIST_ListCount,
		&_LIST_BtreeFirst,
		&_LIST_BtreeLast,
		&_LIST_BtreeNext,
		&_LIST_BtreePrevious,
		&_LIST_CloneList,
		&_LIST_Destroy,
		&_LIST_NewNode,
		&BTREE_Add,
		&BTREE_After,
		&_LIST_BtreeRemove,
		&_LIST_BtreeFind,
		LIST_TYPE_BTREE,
		&_LISTbtree,
#ifdef LISTDEBUG
		"Binary Tree"
#endif /* LISTDEBUG */
		};

/*
 *	This is the default settings for a new
 *	object of this type.
 */
struct _listclass _LISTbtree =
		{
		&master,					/* Pointer to master class data */
		sizeof(struct _btreeh), 	/* Size of the list header */
		sizeof(struct _btree),		/* Size of the node header */
		0,							/* Size of user node data */
		0,							/* Clone count */
		LIST_DUPLICATES_NONE,		/* Duplicate Control */
		NULL,						/* User node Initialise */
		NULL,						/* User node Destroy */
		NULL,						/* user node Compare */
		};
